//
// See README-LCALS_license.txt for access and distribution restrictions
//

//
// Header file with macros and constants for data types, execution,
// timing options, etc. used in LCALS
//

#ifndef LCALSParams_HXX
#define LCALSParams_HXX


////////////////////////////////////////////////////////////////////////////////
//
// This file contains various parameters that control compilation and
// execution of the loop suite.  The macro constants and typedefs in
// this file provide the ability to make changes that will propagate 
// throughout the LCALS code when compiled.  Parameters in this file specify:
//
// o Timing and checksum output
// o Scalar data types and pointer types (e.g., restrict & alignment properties)
// o Loop variants to execute
// o Loop execution policies for traversal templates (used with loop bodies
//   represented as lambda expressions or functors)
//
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// 
// Macro constants that control checksum and timing output.
// Each one should be either defined oor undefined.
// 
// These mean the following:
//
//   - VERIFY_CHECKSUM --> Verify that all variants of each loop produce 
//     numerical results.  When checksum verification is enabled,
//     all data used by each loop is re-initialized before each loop is run.  
//     This adds to overall execution time but not individual loop timing.
//     Also, the checksums are reset with each pass through the loop suite
//     to prevent inconsistencies across runs with different suite pass
//     counts and overflowing the checksum values.  However, some loop 
//     checksums depend on the number of times the loop is sampled.  So,
//     it is possible to have different checksum values for runs with 
//     different loop sample counts.
//
//   - VERIFY_CHECKSUM_ABBREVIATED --> Same as above but uses and abbreviated 
//     run where the number of passes through the suite is one and loop 
//     sampling is cut back to value set below to reduce run time.  When 
//     this macro constant is defined, it supercedes VERIFY_CHECKSUM.  
//
//   - USE_CYCLE --> Use mechanism in cycle.h for loop timing. 
//     Else, use clock mechanism in standard library time.h file.
//
//     Cycle is a very high resolution way to generate an abstract run time.
//     The execution timing data in this case cannot be translated easily
//     into seconds of run time.  Thus, it should only be used to compute
//     relative execution times.
//
//     Clock uses standard clock ticks which can be used to compute seconds
//     of execution time.  It has a coarser grain resolution to the cycle
//     option, but is usually adequate if loop execution requires a sufficient
//     amount of execution time.
//
////////////////////////////////////////////////////////////////////////////////


#define VERIFY_CHECKSUM
//#undef VERIFY_CHECKSUM

//#define VERIFY_CHECKSUM_ABBREVIATED
#undef VERIFY_CHECKSUM_ABBREVIATED

//#define USE_CYCLE
#undef USE_CYCLE


#if defined(VERIFY_CHECKSUM) || defined(VERIFY_CHECKSUM_ABBREVIATED) 
//
// Set define so all loop data will be reinitialized before execution of 
// each loop. This insures that checksums will be the same if arithmetic is. 
//
#define VERIFY_CHECKSUM
#define INIT_DATA
#else
#undef INIT_DATA
#endif

#if defined(VERIFY_CHECKSUM_ABBREVIATED)
static const int num_checksum_suite_passes = 1;
static const int num_checksum_samples = 3;
#endif


////////////////////////////////////////////////////////////////////////////////
//
// Define/undefine macro constants and other paramters used to control data
// type definitions.
// 
////////////////////////////////////////////////////////////////////////////////

//
//  Parameterized scalar data types.
//
#include<complex>

typedef double  Real_type;
typedef std::complex<Real_type> Complex_type;
typedef int     Indx_type;
//
// Use volatile keyword on loop variable for sampling loops to prevent 
// compilers from potentially optimizing out loops where result is
// identical for each sample iteration.
// 
typedef volatile int SampIndx_type;
//
// Use unsigned long for loop variable used in loops to flush cache to
// allow for large caches with size possibly bigger than what int can adderss.
// 
typedef unsigned long CacheIndx_type;

//
//  Floating point array data alignmnent value.  Typically, same as
//  SIMD vector width.
//
const int LCALS_DATA_ALIGN = 32;


////////////////////////////////////////////////////////////////////////////////
//
// Defines for data pointer attributes.  Definitions can be found below.
//
// IMPORTANT: One and only one of these macro constants must be defined!!!
// 
////////////////////////////////////////////////////////////////////////////////
//#define USE_BARE_PTR
#define USE_RESTRICT_PTR
//#define USE_RESTRICT_ALIGNED_PTR
//#define USE_PTR_CLASS
////////////////////////////////////////////////////////////////////////////////


#if defined(COMPILER_XLC12) 
//
// This is here to check whether XLC compiler will SIMD vectorize loops with
// __alignx intrinsic calls in them...
// 
//#define XLC_ALIGN_DECORATE
#undef XLC_ALIGN_DECORATE  

extern
#ifdef __cplusplus
"builtin"
#endif
void __alignx(int n, const void* addr);

#endif

#if defined(COMPILER_ICC)
// alignment attribute supported for versions > 12
#if __ICC >= 1300
typedef Real_type* __restrict__ __attribute__((align_value(LCALS_DATA_ALIGN))) TDRAReal_ptr;
#endif
#endif

#if defined(COMPILER_CLANG)
typedef Real_type aligned_real_type __attribute__((aligned (LCALS_DATA_ALIGN)));
typedef aligned_real_type* __restrict__ TDRAReal_ptr;
#endif

#if defined(USE_PTR_CLASS)
//
//  Class representing a restricted aligned Real_type pointer
//
class RestrictAlignRealPtr
{
public:

   RestrictAlignRealPtr() : dptr(0) { ; }

   RestrictAlignRealPtr(Real_type* d) : dptr(d) { ; }

   //
   // Note: bracket operator implementation is compiler-specific.
   //

#if defined(COMPILER_GNU)
   Real_type& operator [] (Indx_type i)
   {
      return( (Real_type* __restrict__) __builtin_assume_aligned( dptr, LCALS_DATA_ALIGN ) )[i];
   }
#endif

#if defined(COMPILER_ICC)
   Real_type& operator [] (Indx_type i)
   {
#if __ICC < 1300 // use alignment intrinsic
      __assume_aligned( dptr, LCALS_DATA_ALIGN );
      return( (Real_type* __restrict__) dptr)[i];
#else // use alignment attribute
      return( (TDRAReal_ptr) dptr)[i];
#endif
   }
#endif

#if defined(COMPILER_XLC12)
   Real_type& operator [] (Indx_type i)
   {
      __alignx( LCALS_DATA_ALIGN, dptr );
      return( (Real_type* __restrict__) dptr)[i];
   }
#endif

#if defined(COMPILER_CLANG)
   Real_type& operator [] (Indx_type i)
   {
      return( (TDRAReal_ptr) dptr)[i];
   }
#endif

   //
   // This was avoided since it forces operator[] to be defined.
   //
// operator Real_type*() { return dptr; }
   Real_type* get() { return dptr; }

   Real_type* operator+ (Indx_type i) { return dptr+i; }

private:
   Real_type* dptr;
};


//
//  Class representing a restricted Complex_type pointer
//
class RestrictComplexPtr
{
public:

   RestrictComplexPtr() : dptr(0) { ; }

   RestrictComplexPtr(Complex_type* d) : dptr(d) { ; }

   Complex_type& operator [] (Indx_type i)
   {
      return( (Complex_type* __restrict__) dptr)[i];
   }

   //
   // This was avoided since it forces operator[] to be defined.
   //
// operator Complex_type*() { return dptr; }
   Complex_type* get() { return dptr; }

   Complex_type* operator+ (Indx_type i) { return dptr+i; }

private:
   Complex_type* dptr;
};
#endif  // defined(USE_PTR_CLASS)


#if defined(USE_BARE_PTR)
typedef Real_type* Real_ptr;
typedef Complex_type* Complex_ptr;
#endif

#if defined(USE_RESTRICT_PTR)
typedef Real_type* __restrict__ Real_ptr;
typedef Complex_type* __restrict__ Complex_ptr;
#endif

#if defined(USE_RESTRICT_ALIGNED_PTR)
typedef TDRAReal_ptr Real_ptr;
typedef Complex_type* __restrict__ Complex_ptr;
#endif

#if defined(USE_PTR_CLASS)
typedef RestrictAlignRealPtr Real_ptr;
typedef RestrictComplexPtr Complex_ptr;
#endif



////////////////////////////////////////////////////////////////////////////////
//
// Compiler-specific definitions, including which variants can be run for a
// given compiler.  
//
// NOTE: The Makefile sets the COMPILER_* macro constant.
//
//  --> IMPORTANT: There are macro constants defined at the end of this file to
//                 override default loop variants to run.
//
////////////////////////////////////////////////////////////////////////////////

#if defined(COMPILER_ICC)
//
// Configuration options for Intel compilers
//

#define INLINE_DIRECTIVE inline  __attribute__((always_inline))

#define DO_RAW_VARIANTS
#define DO_LAMBDA_VARIANTS
#define DO_FUNCTOR_VARIANTS
#define DO_OMP_VARIANTS

#endif   // end  Intel compilers.....


#if defined(COMPILER_GNU)
//
// Configuration options for GNU compilers
//

#define INLINE_DIRECTIVE inline  __attribute__((always_inline))

#define DO_RAW_VARIANTS
#define DO_LAMBDA_VARIANTS
#define DO_FUNCTOR_VARIANTS
#define DO_OMP_VARIANTS

#endif   // end  GNU compilers.....


#if defined(COMPILER_XLC9) || defined(COMPILER_XLC12)
//
// Configuration options for IBM xlC compilers
//

#define INLINE_DIRECTIVE inline

//
// xlC compilers DO NOT support lambda functions currently!!
//
#define DO_RAW_VARIANTS
#undef DO_LAMBDA_VARIANTS
#define DO_FUNCTOR_VARIANTS
#define DO_OMP_VARIANTS

#endif   // end  XLC compilers.....


#if defined(COMPILER_CLANG)
//
// Configuration options for clang compilers
//

#define INLINE_DIRECTIVE inline  __attribute__((always_inline))

//
// Clang compilers DO NOT support OpenMP currently!!
//
#define DO_RAW_VARIANTS
#define DO_LAMBDA_VARIANTS
#define DO_FUNCTOR_VARIANTS
#undef DO_OMP_VARIANTS

#endif   // end  GNU compilers.....



//
// By default, all loop variants defined as supported in the
// compiler-specific sections above are turned on here (for
// both compilation and execution).
//
// Loop variants can be turned off via the #define/#undef macros below.
// It's a bit cheesy, but it's simple and it works!
//
// Execution of individual loop variants can also be controlled
// by modifying the strings that get added to the run_loop_variants
// vector in main.cxx
//

////////////////////////////////////////////////////////////////////////////////
//
// IMPORTANT: Many variants have been built for various experiments and
//            are not available in the generally available version of LCALS.
//
////////////////////////////////////////////////////////////////////////////////

//#undef DO_RAW_VARIANTS
//#undef DO_OMP_VARIANTS
//#undef DO_LAMBDA_VARIANTS
#undef DO_FUNCTOR_VARIANTS


#if defined(DO_RAW_VARIANTS)
#define DO_RAW
//#define DO_RAW_FUNC
#endif

#if defined(DO_RAW_VARIANTS) && defined(DO_OMP_VARIANTS)
#define DO_RAW_OMP
#endif

#if defined(DO_LAMBDA_VARIANTS)
#define DO_LAMBDA
//#define DO_LAMBDA_TYPEFIX
//#define DO_HYBRID_LAMBDA
//#define DO_HYBRID_LAMBDA_TYPEFIX
#endif

#if defined(DO_LAMBDA_VARIANTS) && defined(DO_OMP_VARIANTS) 
#define DO_LAMBDA_OMP
//#define DO_LAMBDA_OMP_TYPEFIX
//#define DO_HYBRID_LAMBDA_OMP  not working yet!
#endif

#if defined(DO_FUNCTOR_VARIANTS)
#define DO_FUNCTOR
//#define DO_HYBRID_FUNCTOR  not working yet!
#endif

#if defined(DO_FUNCTOR_VARIANTS) && defined(DO_OMP_VARIANTS) 
#define DO_FUNCTOR_OMP
//#define DO_HYBRID_LAMBDA_OMP  not working yet!
#endif

//
//  OMP_PRIVATEVARS  controls the use of the firstprivate clause for 
//  read-only variables in OpenMP Raw loop variants.
//
//#define OMP_PRIVATEVARS
#undef OMP_PRIVATEVARS


//
//  Traversal policies applicable to "forall" loop variants.  
//  Traversal method templates are defined in LCALSTraversalMethods.hxx 
//  header files.
//

// Tag struct types define available forall method traversal policies
struct seq_traversal {};
struct simd_traversal {};
struct omp_parallel_for_traversal {};
struct omp_for_nowait_traversal {};

// 
// Traversal policy used in (non-OpenMP) "forall" loop variants.  
// To use another policy in all such loops, change this typedef.
//
typedef simd_traversal traversal_policy;



#endif  // closing endif for header file include guard
