//
// See README-LCALS_license.txt for access and distribution restrictions
//

//
// Source file with routines to allocate data for LCALS suite
// and define parameters controlling execution of each loop.
//

#include "LCALSSuite.hxx"
#include "LCALSStats.hxx"

#include "SubsetDataA.hxx"

#include<cstdlib>
#include<string>
#include<iostream>

#include<sys/types.h>
#include<sys/stat.h>


//
// File scope data holding structures used in loop suite
//
static LoopData* s_loop_data = 0;


//
// Prototypes for file scope routines used in to manage loop data and checksums
//

namespace {

Real_ptr allocAndInitData(LoopData::RealArray& ra, Indx_type len);
Indx_type* allocAndInitData(LoopData::IndxArray& ia, Indx_type len);
Complex_ptr allocAndInitData(LoopData::ComplexArray& ca, Indx_type len);
void initData(LoopData::RealArray& ra);
void initData(LoopData::IndxArray& ia);
void initData(LoopData::ComplexArray& ca);

void initChksum(LoopStat& stat, LoopLength ilength);
void updateChksum(LoopStat& stat, LoopLength ilength,
                  const LoopData::RealArray& ra, Real_type scale_factor = 1.0);
void updateChksum(LoopStat& stat, LoopLength ilength,
                  Real_type val);
void updateChksum(LoopStat& stat, LoopLength ilength,
                  const LoopData::ComplexArray& ca, Real_type scale_factor = 1.0);

}  // closing brace for unnamed namespace




//
// Accessor routine for suite kernel data.
//
LoopData& getLoopData() { return *s_loop_data; }



//
// Define how suite will run and initialize stat structures for loops.
//
// NOTE: Loop lengths, loop sample counts (and weights for optimization
//       evaluation) are defined here!
//
// These values should be set large enough to accurately generate 
// execution timings (i.e., not too small to be masked by CPU timing 
// resolution and overhead).  The values set here were manually determined 
// so that O(1) seconds of execution time is required to sample each loop 
// on some of our fastest Intel machines.
//
void defineLoopSuiteRunInfo(bool run_loop[], 
                            std::vector< std::string >& run_loop_variants,
                            double sample_frac)
{
   std::cout << "\n defineLoopSuiteRunInfo..." << std::endl;

   if ( s_loop_data == 0 ) { 
      s_loop_data = new LoopData(); 
   }

   //
   //
   // Enumeration defining loop groups for relative weighting of
   // execution timing based on what we think is most important.
   //
   // In computation of figures of merit (FOM), loops with higher 
   // weights will reduce FOM value more for higher run-time than 
   // those with lower weights.
   //
   enum WeightGroup {

      DATA_PARALLEL = 0,
      ORDER_DEPENDENT,
      TRANSCENDENTAL,
      DATA_DEPENDENT,
      POINTER_NEST,
      COMPLEX,

      NUM_WEIGHT_GROUPS  // Keep this one last and NEVER comment out (!!)
   };


   //
   // Initialize structure holding loop suite execution data.
   //
   LoopSuiteRunInfo& suite_info = getLoopSuiteRunInfo();

   suite_info.loop_samp_frac = sample_frac;

   suite_info.loop_weights.resize(NUM_WEIGHT_GROUPS);
   suite_info.loop_weights[DATA_PARALLEL]   = 2.0;
   suite_info.loop_weights[ORDER_DEPENDENT] = 1.8;
   suite_info.loop_weights[TRANSCENDENTAL]  = 1.7;
   suite_info.loop_weights[DATA_DEPENDENT]  = 1.7;
   suite_info.loop_weights[POINTER_NEST]    = 1.4;
   suite_info.loop_weights[COMPLEX]         = 1.0;

   suite_info.loop_length_names.resize(NUM_LENGTHS);
   suite_info.loop_length_names[LONG] = std::string("LONG");
   suite_info.loop_length_names[MEDIUM] = std::string("MEDIUM");
   suite_info.loop_length_names[SHORT] = std::string("SHORT");

   suite_info.num_loops_run.resize( run_loop_variants.size() );
   suite_info.tot_time.resize( run_loop_variants.size() );
   suite_info.fom_rel.resize( run_loop_variants.size() );
   suite_info.fom_rate.resize( run_loop_variants.size() );

   for (unsigned ilv = 0; ilv < run_loop_variants.size(); ++ilv) {
      suite_info.addLoopStats(run_loop_variants[ilv]);

      suite_info.num_loops_run[ilv].resize(NUM_LENGTHS, 0);
      suite_info.tot_time[ilv].resize(NUM_LENGTHS, 0.0);
      suite_info.fom_rel[ilv].resize(NUM_LENGTHS, 0.0);
      suite_info.fom_rate[ilv].resize(NUM_LENGTHS, 0.0);
   }


   std::vector<double>& weight = suite_info.loop_weights;

   Indx_type max_loop_length = 0;

   for (unsigned iloop = 0 ; iloop < suite_info.num_loops; ++iloop) {

      std::string loop_name;
      LoopStat loop_stat(suite_info.num_loop_lengths);

      Indx_type max_loop_indx = 0;

      if ( run_loop[iloop] ) {

         switch ( iloop ) {
   
            case REF_LOOP : {
               loop_name = std::string("REF_LOOP"); 
               //
               // Note: Reference loop stats are not used in
               //       in suite.  Parameters are defined in 
               //       defineReferenceLoopRunInfo( ) routine.
               //
               break;
            }


            //
            // Parameters defining how loops in Subset A are run...
            //
            case PRESSURE_CALC   : {
               loop_name = std::string("PRESSURE_CALC");

               loop_stat.loop_weight = weight[DATA_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 20000;
               loop_stat.samples_per_pass[MEDIUM] = 200000;
               loop_stat.samples_per_pass[SHORT]  = 20000000;

               break;
            }

            case ENERGY_CALC   : {
               loop_name = std::string("ENERGY_CALC");

               loop_stat.loop_weight = weight[DATA_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 5000;
               loop_stat.samples_per_pass[MEDIUM] = 40000;
               loop_stat.samples_per_pass[SHORT]  = 2000000;

               break;
            }

            case VOL3D_CALC   : {
               loop_name = std::string("VOL3D_CALC");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               Indx_type ndims = 3;

               ADomain Ldomain(LONG, ndims);
               loop_stat.loop_length[LONG]   = Ldomain.lpz - Ldomain.fpz + 1;
               ADomain Mdomain(MEDIUM, ndims);
               loop_stat.loop_length[MEDIUM] = Mdomain.lpz - Mdomain.fpz + 1;
               ADomain Sdomain(SHORT, ndims);
               loop_stat.loop_length[SHORT]  = Sdomain.lpz - Sdomain.fpz + 1;

               max_loop_indx = Ldomain.lpn;

               loop_stat.samples_per_pass[LONG]   = 6000;
               loop_stat.samples_per_pass[MEDIUM] = 30000;
               loop_stat.samples_per_pass[SHORT]  = 800000;

               break;
            }

            case DEL_DOT_VEC_2D : {
               loop_name = std::string("DEL_DOT_VEC_2D");

               loop_stat.loop_weight = weight[DATA_PARALLEL]; 

               Indx_type ndims = 2;

               ADomain Ldomain(LONG, ndims);
               loop_stat.loop_length[LONG]   = Ldomain.n_real_zones;
               ADomain Mdomain(MEDIUM, ndims);
               loop_stat.loop_length[MEDIUM] = Mdomain.n_real_zones;
               ADomain Sdomain(SHORT, ndims);
               loop_stat.loop_length[SHORT]  = Sdomain.n_real_zones;

               max_loop_indx = Ldomain.lrn;

               loop_stat.samples_per_pass[LONG]   = 4000;
               loop_stat.samples_per_pass[MEDIUM] = 25000;
               loop_stat.samples_per_pass[SHORT]  = 2000000;

               break;
            }

            case COUPLE   : {
               loop_name = std::string("COUPLE");

               loop_stat.loop_weight = weight[TRANSCENDENTAL]; 

               Indx_type ndims = 3;

               ADomain Ldomain(LONG, ndims);
               loop_stat.loop_length[LONG]   = Ldomain.lpz - Ldomain.fpz + 1;
               ADomain Mdomain(MEDIUM, ndims);
               loop_stat.loop_length[MEDIUM] = Mdomain.lpz - Mdomain.fpz + 1;
               ADomain Sdomain(SHORT, ndims);
               loop_stat.loop_length[SHORT]  = Sdomain.lpz - Sdomain.fpz + 1;

               max_loop_indx = Ldomain.lrn;

               loop_stat.samples_per_pass[LONG]   = 1600;
               loop_stat.samples_per_pass[MEDIUM] = 10000;
               loop_stat.samples_per_pass[SHORT]  = 600000;

               break;
            }

            case FIR   : {
               loop_name = std::string("FIR");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 12000;
               loop_stat.samples_per_pass[MEDIUM] = 90000;
               loop_stat.samples_per_pass[SHORT]  = 6000000;

               break;
            }


            //
            // Parameters defining how loops in Subset B are run...
            //
            case INIT3 : {
               loop_name = std::string("INIT3");

               loop_stat.loop_weight = weight[DATA_PARALLEL]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 16000;
               loop_stat.samples_per_pass[MEDIUM] = 130000;
               loop_stat.samples_per_pass[SHORT]  = 25000000;

               break;
            }

            case MULADDSUB : { 
               loop_name = std::string("MULADDSUB"); 

               loop_stat.loop_weight = weight[DATA_PARALLEL]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 16000;
               loop_stat.samples_per_pass[MEDIUM] = 140000;
               loop_stat.samples_per_pass[SHORT]  = 30000000;

               break; 
            }

            case IF_QUAD : {
               loop_name = std::string("IF_QUAD");

               loop_stat.loop_weight = weight[DATA_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 5000;
               loop_stat.samples_per_pass[MEDIUM] = 35000;
               loop_stat.samples_per_pass[SHORT]  = 1800000;

               break;
            }

            case TRAP_INT : {
               loop_name = std::string("TRAP_INT");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 6000;
               loop_stat.samples_per_pass[MEDIUM] = 40000;
               loop_stat.samples_per_pass[SHORT]  = 2500000;

               break;
            }


            //
            // Parameters defining how loops in Subset C are run...
            //
            case HYDRO_1D : { 
               loop_name = std::string("HYDRO_1D"); 

               loop_stat.loop_weight = weight[DATA_PARALLEL]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 50000;
               loop_stat.samples_per_pass[MEDIUM] = 400000;
               loop_stat.samples_per_pass[SHORT]  = 20000000;

               break; 
            }

            case ICCG : {
               loop_name = std::string("ICCG"); 

               loop_stat.loop_weight = weight[COMPLEX]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 30000;
               loop_stat.samples_per_pass[MEDIUM] = 200000;
               loop_stat.samples_per_pass[SHORT]  = 10000000;

               break; 
            }

            case INNER_PROD : {
               loop_name = std::string("INNER_PROD"); 

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 80000;
               loop_stat.samples_per_pass[MEDIUM] = 600000;
               loop_stat.samples_per_pass[SHORT]  = 50000000;

               break; 
            }

            case BAND_LIN_EQ : {
               loop_name = std::string("BAND_LIN_EQ"); 

               loop_stat.loop_weight = weight[COMPLEX]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 80000;
               loop_stat.samples_per_pass[MEDIUM] = 800000;
               loop_stat.samples_per_pass[SHORT]  = 30000000;

               break; 
            }

            case TRIDIAG_ELIM : {
               loop_name = std::string("TRIDIAG_ELIM"); 

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 20000;
               loop_stat.samples_per_pass[MEDIUM] = 100000;
               loop_stat.samples_per_pass[SHORT]  = 75000000;

               break; 
            }

            case EOS   : { 
               loop_name = std::string("EOS");

               loop_stat.loop_weight = weight[DATA_PARALLEL]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 24000;
               loop_stat.samples_per_pass[MEDIUM] = 160000;
               loop_stat.samples_per_pass[SHORT]  = 9000000;
            
               break; 
            }

            case ADI   : {
               loop_name = std::string("ADI");

               loop_stat.loop_weight = weight[COMPLEX]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 1600;
               loop_stat.samples_per_pass[MEDIUM] = 9000;
               loop_stat.samples_per_pass[SHORT]  = 650000;
            
               break; 
            }

            case INT_PREDICT   : {
               loop_name = std::string("INT_PREDICT");

               loop_stat.loop_weight = weight[POINTER_NEST]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 5600;
               loop_stat.samples_per_pass[MEDIUM] = 40000;
               loop_stat.samples_per_pass[SHORT]  = 3400000;
            
               break; 
            }

            case DIFF_PREDICT   : {
               loop_name = std::string("DIFF_PREDICT");

               loop_stat.loop_weight = weight[POINTER_NEST]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 3500;
               loop_stat.samples_per_pass[MEDIUM] = 25000;
               loop_stat.samples_per_pass[SHORT]  = 2400000;
            
               break; 
            }

            case FIRST_SUM   : { 
               loop_name = std::string("FIRST_SUM");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 50000;
               loop_stat.samples_per_pass[MEDIUM] = 280000;
               loop_stat.samples_per_pass[SHORT]  = 20000000;

               break; 
            }

            case FIRST_DIFF   : { 
               loop_name = std::string("FIRST_DIFF");

               loop_stat.loop_weight = weight[DATA_PARALLEL]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 50000;
               loop_stat.samples_per_pass[MEDIUM] = 500000;
               loop_stat.samples_per_pass[SHORT]  = 50000000;

               break; 
            }

            case PIC_2D   : {
               loop_name = std::string("PIC_2D");

               loop_stat.loop_weight = weight[COMPLEX]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 3000;
               loop_stat.samples_per_pass[MEDIUM] = 18000;
               loop_stat.samples_per_pass[SHORT]  = 1800000;

               break; 
            }

            case PIC_1D   : {
               loop_name = std::string("PIC_1D");

               loop_stat.loop_weight = weight[DATA_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 3300;
               loop_stat.samples_per_pass[MEDIUM] = 24000;
               loop_stat.samples_per_pass[SHORT]  = 2000000;

               break; 
            }

            case HYDRO_2D   : {   
               loop_name = std::string("HYDRO_2D");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 700;
               loop_stat.samples_per_pass[MEDIUM] = 4000;
               loop_stat.samples_per_pass[SHORT]  = 250000;

               break; 
            }

            case GEN_LIN_RECUR   : { 
               loop_name = std::string("GEN_LIN_RECUR");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 6000;
               loop_stat.samples_per_pass[MEDIUM] = 36000;
               loop_stat.samples_per_pass[SHORT]  = 2000000;

               break; 
            }

            case DISC_ORD   : { 
               loop_name = std::string("DISC_ORD");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 1800;
               loop_stat.samples_per_pass[MEDIUM] = 12000;
               loop_stat.samples_per_pass[SHORT]  = 700000;

               break; 
            }

            case MAT_X_MAT   : {   
               loop_name = std::string("MAT_X_MAT");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 12;
               loop_stat.samples_per_pass[MEDIUM] = 100;
               loop_stat.samples_per_pass[SHORT]  = 15000;

               break; 
            }

            case PLANCKIAN   : { 
               loop_name = std::string("PLANCKIAN");

               loop_stat.loop_weight = weight[TRANSCENDENTAL]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 6500;
               loop_stat.samples_per_pass[MEDIUM] = 42000;
               loop_stat.samples_per_pass[SHORT]  = 2400000;

               break; 
            }

            case IMP_HYDRO_2D   : {
               loop_name = std::string("IMP_HYDRO_2D");

               loop_stat.loop_weight = weight[ORDER_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 2000;
               loop_stat.samples_per_pass[MEDIUM] = 13000;
               loop_stat.samples_per_pass[SHORT]  = 600000;

               break; 
            }

            case FIND_FIRST_MIN   : { 
               loop_name = std::string("FIND_FIRST_MIN");

               loop_stat.loop_weight = weight[DATA_DEPENDENT]; 

               loop_stat.loop_length[LONG]   = 25023;
               loop_stat.loop_length[MEDIUM] = 4007;
               loop_stat.loop_length[SHORT]  = 67;

               max_loop_indx = loop_stat.loop_length[LONG];

               loop_stat.samples_per_pass[LONG]   = 55000;
               loop_stat.samples_per_pass[MEDIUM] = 330000;
               loop_stat.samples_per_pass[SHORT]  = 14000000;

               break; 
            }


            default : { 
               std::cout << "\n Unknown loop id = " << iloop << std::endl; 
            }

         }  // switch statement on loop id

      }  // if loop with id is to be run


      suite_info.loop_names.push_back(loop_name); 

      //
      // Set max loop length to be largest loop index used over all loops.
      // 
      max_loop_length = 
         std::max(max_loop_length, max_loop_indx);
            
      //
      // Set number of times each loop length will be run.
      // 
      for (unsigned i = 0; i < suite_info.num_loop_lengths; ++i) {

         loop_stat.samples_per_pass[i] = static_cast<int>(
            loop_stat.samples_per_pass[i] * suite_info.loop_samp_frac ); 

         if ( suite_info.run_loop_length[i] ) {
            loop_stat.loop_run_count[i] = 
               loop_stat.samples_per_pass[i] * suite_info.num_suite_passes;
         } else {
            loop_stat.loop_run_count[i] = 0;
         }

      }

      //
      // We add loop stat for each loop to maintain consistent array indexing.
      // However, only loops specified to be run will be executed.
      // 
      for (unsigned ilv = 0; ilv < run_loop_variants.size(); ++ilv) {
         suite_info.getLoopStats(run_loop_variants[ilv]).push_back(loop_stat);
      }

   }  // loop over loop IDs


   defineReferenceLoopRunInfo();

   s_loop_data->max_loop_length = 
      std::max(max_loop_length, suite_info.ref_loop_stat.loop_length[LONG]); 
}



//
// Initialize data to run loop with given ID.  Note that this routine 
// assumes that it is called before the loop with given ID is run and
// that data initialization calls in here are concistent with what is
// needed to execute loop.
//
// Loop data is initialized in this routine so all variants of loop
// tun the same way.  Note that data arrays are initialized for 
// each loop only under the circumstances that it is actually required.
//
void loopInit(unsigned iloop, LoopStat& stat)
{
   LoopData& loop_data = getLoopData();

   flushCache();

   stat.loop_is_run = true;


   switch ( iloop ) {
   
      case REF_LOOP : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);

         break;
      }


      //
      // Initialize data for Loop Subset A...
      //
      case PRESSURE_CALC   : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);

         initData(loop_data.RealArray_scalars); 

         break;
      }

      case ENERGY_CALC   : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);
         initData(loop_data.RealArray_1D[5]);
         initData(loop_data.RealArray_1D[6]);
         initData(loop_data.RealArray_1D[7]);
         initData(loop_data.RealArray_1D[8]);
         initData(loop_data.RealArray_1D[9]);
         initData(loop_data.RealArray_1D[10]);
         initData(loop_data.RealArray_1D[11]);
         initData(loop_data.RealArray_1D[12]);
         initData(loop_data.RealArray_1D[13]);
         initData(loop_data.RealArray_1D[14]);

         initData(loop_data.RealArray_scalars);

         break;
      }

      case VOL3D_CALC   : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);

         break;
      }

      case DEL_DOT_VEC_2D : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);

         break;
      }

      case COUPLE   : {

         initData(loop_data.ComplexArray_1D[0]);
         initData(loop_data.ComplexArray_1D[1]);
         initData(loop_data.ComplexArray_1D[2]);
         initData(loop_data.ComplexArray_1D[3]);
         initData(loop_data.ComplexArray_1D[4]);

         break;
      }

      case FIR   : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);

         break;
      }


      //
      // Initialize data for Loop Subset B...
      //
      case INIT3 : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);
#endif

         break;
      }

      case MULADDSUB : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);
#endif

         break; 
      }

      case IF_QUAD : {

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);

         break;
      }

      case TRAP_INT : {

         initData(loop_data.IndxArray_1D[0]);

         initData(loop_data.RealArray_scalars);

         break;
      }


      //
      // Initialize data for Loop Subset C...
      //
      case HYDRO_1D : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);

         initData(loop_data.RealArray_scalars);
#endif

         break; 
      }

      case ICCG : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D_Nx4[0]);
         initData(loop_data.RealArray_1D_Nx4[1]);
#endif

         break; 
      }

      case INNER_PROD : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
#endif

         break; 
      }

      case BAND_LIN_EQ : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
#endif

         break; 
      }

      case TRIDIAG_ELIM : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
#endif

         break; 
      }

      case EOS   : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);

         initData(loop_data.RealArray_scalars);
#endif
      
         break; 
      }

      case ADI   : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);

         initData(loop_data.RealArray_3D_2xNx4[0]);
         initData(loop_data.RealArray_3D_2xNx4[1]);
         initData(loop_data.RealArray_3D_2xNx4[2]);

         initData(loop_data.RealArray_scalars);
#endif
      
         break; 
      }

      case INT_PREDICT   : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_2D_Nx25[0]);

         initData(loop_data.RealArray_scalars);
#endif
      
         break; 
      }

      case DIFF_PREDICT   : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_2D_Nx25[0]);
         initData(loop_data.RealArray_2D_Nx25[1]);
#endif
      
         break; 
      }

      case FIRST_SUM   : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
#endif

         break; 
      }

      case FIRST_DIFF   : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
#endif

         break; 
      }

      case PIC_2D   : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_2D_Nx25[0]);
         initData(loop_data.RealArray_2D_Nx25[1]);
         initData(loop_data.RealArray_2D_Nx25[2]);

         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);

         initData(loop_data.IndxArray_1D[0]);
         initData(loop_data.IndxArray_1D[1]);

         initData(loop_data.RealArray_2D_64x64[0]);
#endif

         break; 
      }

      case PIC_1D   : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);
         initData(loop_data.RealArray_1D[5]);
         initData(loop_data.RealArray_1D[6]);
         initData(loop_data.RealArray_1D[7]);
         initData(loop_data.RealArray_1D[8]);

         initData(loop_data.RealArray_scalars);

         initData(loop_data.IndxArray_1D[2]);
         initData(loop_data.IndxArray_1D[3]);
         initData(loop_data.IndxArray_1D[4]);
#endif

         break; 
      }

      case HYDRO_2D   : {   

#if defined(INIT_DATA)
         initData(loop_data.RealArray_2D_7xN[0]);
         initData(loop_data.RealArray_2D_7xN[1]);
         initData(loop_data.RealArray_2D_7xN[2]);
         initData(loop_data.RealArray_2D_7xN[3]);
         initData(loop_data.RealArray_2D_7xN[4]);
         initData(loop_data.RealArray_2D_7xN[5]);
         initData(loop_data.RealArray_2D_7xN[6]);
         initData(loop_data.RealArray_2D_7xN[7]);
         initData(loop_data.RealArray_2D_7xN[8]);
         initData(loop_data.RealArray_2D_7xN[9]);
         initData(loop_data.RealArray_2D_7xN[10]);
#endif

         break; 
      }

      case GEN_LIN_RECUR   : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);

         initData(loop_data.RealArray_scalars);
#endif

         break; 
      }

      case DISC_ORD   : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);
         initData(loop_data.RealArray_1D[5]);
         initData(loop_data.RealArray_1D[6]);
         initData(loop_data.RealArray_1D[7]);
         initData(loop_data.RealArray_1D[8]);

         initData(loop_data.RealArray_scalars);
#endif

         break; 
      }

      case MAT_X_MAT   : {   

#if defined(INIT_DATA)
         initData(loop_data.RealArray_2D_Nx25[0]);
         initData(loop_data.RealArray_2D_Nx25[1]);
         initData(loop_data.RealArray_2D_64x64[0]);
#endif

         break; 
      }

      case PLANCKIAN   : { 

#if defined(INIT_DATA)
         initData(loop_data.RealArray_1D[0]);
         initData(loop_data.RealArray_1D[1]);
         initData(loop_data.RealArray_1D[2]);
         initData(loop_data.RealArray_1D[3]);
         initData(loop_data.RealArray_1D[4]);
#endif

         break; 
      }

      case IMP_HYDRO_2D   : {

#if defined(INIT_DATA)
         initData(loop_data.RealArray_2D_7xN[0]);
         initData(loop_data.RealArray_2D_7xN[1]);
         initData(loop_data.RealArray_2D_7xN[2]);
         initData(loop_data.RealArray_2D_7xN[3]);
         initData(loop_data.RealArray_2D_7xN[4]);
         initData(loop_data.RealArray_2D_7xN[5]);
#endif

         break; 
      }

      case FIND_FIRST_MIN   : { 

         initData(loop_data.RealArray_1D[0]);

         break; 
      }


      default : { 
         std::cout << "\n Unknown loop id = " << iloop << std::endl; 
      }

   }  // switch statement on loop id

}


//
// Finalize data for loop with given ID.  Note that this routine assumes 
// that it is called after the loop with given ID is run and that checksum 
// calls in here are concistent with what is needed for loop.
//
void loopFinalize(unsigned iloop, LoopStat& stat, LoopLength ilength)
{
#if defined(VERIFY_CHECKSUM)
   initChksum(stat, ilength);

   LoopData& loop_data = getLoopData();

   switch ( iloop ) {
   
      case REF_LOOP : {

         // Nothing to do for REF_LOOP case...

         break;
      }


      //
      // Update checksums for Loop Subset A...
      //
      case PRESSURE_CALC   : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[2]);

         break;
      }

      case ENERGY_CALC   : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[5]);

         break;
      }

      case VOL3D_CALC   : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[3]);

         break;
      }

      case DEL_DOT_VEC_2D : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[4]);

         break;
      }

      case COUPLE   : {

         updateChksum(stat, ilength, loop_data.ComplexArray_1D[0]);
         updateChksum(stat, ilength, loop_data.ComplexArray_1D[1]);
         updateChksum(stat, ilength, loop_data.ComplexArray_1D[2]);

         break;
      }

      case FIR   : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break;
      }


      //
      // Update checksums for Loop Subset B...
      //
      case INIT3 : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[1]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[2]);

         break;
      }

      case MULADDSUB : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[1]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[2]);

         break; 
      }

      case IF_QUAD : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[3]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[4]);

         break;
      }

      case TRAP_INT : {

         updateChksum(stat, ilength, loop_data.scalar_Real[0]);

         break;
      }


      //
      // Update checksums for Loop Subset C...
      //
      case HYDRO_1D : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break; 
      }

      case ICCG : {

         updateChksum(stat, ilength, loop_data.RealArray_1D_Nx4[0]);

         break; 
      }

      case INNER_PROD : {

         updateChksum(stat, ilength, loop_data.scalar_Real[0]);

         break; 
      }

      case BAND_LIN_EQ : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break; 
      }

      case TRIDIAG_ELIM : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]); 

         break; 
      }

      case EOS   : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break; 
      }

      case ADI   : {

         updateChksum(stat, ilength, loop_data.RealArray_3D_2xNx4[0]);
         updateChksum(stat, ilength, loop_data.RealArray_3D_2xNx4[1]);
         updateChksum(stat, ilength, loop_data.RealArray_3D_2xNx4[2]);

         break; 
      }

      case INT_PREDICT   : {

         updateChksum(stat, ilength, loop_data.RealArray_2D_Nx25[0]);

         break; 
      }

      case DIFF_PREDICT   : {

         updateChksum(stat, ilength, loop_data.RealArray_2D_Nx25[0]);

         break; 
      }

      case FIRST_SUM   : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break; 
      }

      case FIRST_DIFF   : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break; 
      }

      case PIC_2D   : {

         updateChksum(stat, ilength, loop_data.RealArray_2D_Nx25[0]);
         updateChksum(stat, ilength, loop_data.RealArray_2D_64x64[0]); 

         break; 
      }

      case PIC_1D   : {

         updateChksum(stat, ilength, loop_data.RealArray_1D[6]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[1]);
         updateChksum(stat, ilength, loop_data.RealArray_1D[7]);

         break; 
      }

      case HYDRO_2D   : {   

         updateChksum(stat, ilength, loop_data.RealArray_2D_7xN[9]);
         updateChksum(stat, ilength, loop_data.RealArray_2D_7xN[10]);

         break; 
      }

      case GEN_LIN_RECUR   : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[0]);

         break; 
      }

      case DISC_ORD   : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[7]);

         break; 
      }

      case MAT_X_MAT   : {   

         updateChksum(stat, ilength, loop_data.RealArray_2D_Nx25[0]); 

         break; 
      }

      case PLANCKIAN   : { 

         updateChksum(stat, ilength, loop_data.RealArray_1D[4]);

         break; 
      }

      case IMP_HYDRO_2D   : {

         updateChksum(stat, ilength, loop_data.RealArray_2D_7xN[0]);

         break; 
      }

      case FIND_FIRST_MIN   : { 

         updateChksum(stat, ilength, loop_data.scalar_Real[0]); 

         break; 
      }


      default : { 
         std::cout << "\n Unknown loop id = " << iloop << std::endl; 
      }

   }  // switch statement on loop id

#endif  // if VERIFY_CHECKSUM
}


//
// Allocate and initialize arrays (and scalars) used to execute loops in suite.
//
void allocateLoopData()
{
   std::cout << "\n allocateLoopData..." << std::endl;

   unsigned num_aligned_segments = 
      (s_loop_data->max_loop_length + 20)/LCALS_DATA_ALIGN + 1;
   unsigned aligned_chunksize = num_aligned_segments * LCALS_DATA_ALIGN;

   //
   //  Allocate and initialize 1D loop length Real arrays.
   // 
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Real_arrays; ++i) { 
      Indx_type data_len = aligned_chunksize;

      LoopData::RealArray* rarray = s_loop_data->RealArray_1D;
      rarray[i].id = i+1;
      Real_ptr data = allocAndInitData(rarray[i], data_len); 

      s_loop_data->array_1D_Real[i] = data;
   }

   //
   //  Allocate and initialize 1D loop length X 4 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Nx4_Real_arrays; ++i) {
      Indx_type data_len = aligned_chunksize*4;

      LoopData::RealArray* rarray = s_loop_data->RealArray_1D_Nx4;
      rarray[i].id = i+1;
      Real_ptr data = allocAndInitData(rarray[i], data_len);

      s_loop_data->array_1D_Nx4_Real[i] = data;
   }

   //
   //  Allocate and initialize 1D loop length Indx arrays.
   // 
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Indx_arrays; ++i) {
      Indx_type data_len = aligned_chunksize;

      LoopData::IndxArray* iarray = s_loop_data->IndxArray_1D;
      iarray[i].id = i;
      Indx_type* data = allocAndInitData(iarray[i], data_len);

      s_loop_data->array_1D_Indx[i] = data;
   }

   //
   //  Allocate and initialize 1D loop length Complex arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Complex_arrays; ++i) {
      Indx_type data_len = aligned_chunksize;

      LoopData::ComplexArray* carray = s_loop_data->ComplexArray_1D;
      carray[i].id = i+1;
      Complex_ptr data = allocAndInitData(carray[i], data_len);

      s_loop_data->array_1D_Complex[i] = data;
   }

   //
   //  Allocate and initialize 2D loop length X 25 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_2D_Nx25_Real_arrays; ++i) {
      Indx_type data_len = aligned_chunksize*25;

      LoopData::RealArray* rarray = s_loop_data->RealArray_2D_Nx25;
      rarray[i].id = i+1;
      Real_ptr data = allocAndInitData(rarray[i], data_len);

      s_loop_data->array_2D_Nx25_Real[i] = new Real_ptr[aligned_chunksize];
      for (Indx_type k = 0; k < aligned_chunksize; ++k) {
         s_loop_data->array_2D_Nx25_Real[i][k] = &data[k*25];
      }
   }

   //
   //  Allocate and initialize 2D 7 X loop length Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_2D_7xN_Real_arrays; ++i) {
      Indx_type data_len = 7*aligned_chunksize;

      LoopData::RealArray* rarray = s_loop_data->RealArray_2D_7xN;
      rarray[i].id = i+1;
      Real_ptr data = allocAndInitData(rarray[i], data_len);

      s_loop_data->array_2D_7xN_Real[i] = new Real_ptr[7];
      for (Indx_type k = 0; k < 7; ++k) {
         s_loop_data->array_2D_7xN_Real[i][k] = &data[k*aligned_chunksize];
      }
   }

   //
   //  Allocate and initialize 2D 64 X 64 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_2D_64x64_Real_arrays; ++i) {
      Indx_type data_len = 64*64;

      LoopData::RealArray* rarray = s_loop_data->RealArray_2D_64x64;
      rarray[i].id = i+1;
      Real_ptr data = allocAndInitData(rarray[i], data_len);

      s_loop_data->array_2D_64x64_Real[i] = new Real_ptr[64];
      for (Indx_type k = 0; k < 64; ++k) {
         s_loop_data->array_2D_64x64_Real[i][k] = &data[k*64];
      }
   }

   //
   //  Allocate and initialize 3D 2 X loop length X 4 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_3D_2xNx4_Real_arrays; ++i) {
      Indx_type data_len = 2*aligned_chunksize*4;

      LoopData::RealArray* rarray = s_loop_data->RealArray_3D_2xNx4;
      rarray[i].id = i+1;
      Real_ptr data = allocAndInitData(rarray[i], data_len);

      s_loop_data->array_3D_2xNx4_Real[i] = new Real_ptr*[2];
      for (Indx_type k = 0; k < 2; ++k) {
         s_loop_data->array_3D_2xNx4_Real[i][k] = new Real_ptr[aligned_chunksize];
      }

      for (Indx_type k = 0; k < 2; ++k) {
         for (Indx_type l = 0; l < aligned_chunksize; ++l) {
            s_loop_data->array_3D_2xNx4_Real[i][k][l] = &data[k*l*4];
         }
      }
   }


   //
   //  Initialize Real scalars.
   //
   s_loop_data->RealArray_scalars.id = 21.0;
   s_loop_data->RealArray_scalars.data = s_loop_data->scalar_Real;
   s_loop_data->RealArray_scalars.len = s_loop_data->s_num_Real_scalars;
   initData(s_loop_data->RealArray_scalars);

}

//
// Free arrays used in loop suite loop execution (allocated in routine above).
//
void freeLoopData()
{
   if ( s_loop_data != 0 )  return;

   std::cout << "\n freeLoopData..." << std::endl;

   //
   //  De-allocate 1D loop length Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Real_arrays; ++i) {
#if defined(USE_PTR_CLASS)
      free( s_loop_data->array_1D_Real[i].get() );
#else
      free( s_loop_data->array_1D_Real[i] );
#endif
   }

   //
   //  De-allocate 1D loop length X 4 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Nx4_Real_arrays; ++i) {
#if defined(USE_PTR_CLASS)
      free( s_loop_data->array_1D_Nx4_Real[i].get() );
#else
      free( s_loop_data->array_1D_Nx4_Real[i] );
#endif
   }

   //
   //  De-allocate 1D loop length Indx arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Indx_arrays; ++i) {
      free( s_loop_data->array_1D_Indx[i] );
   }

   //
   //  De-allocate 1D loop length Complex arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_1D_Complex_arrays; ++i) {
#if defined(USE_PTR_CLASS)
      free( s_loop_data->array_1D_Complex[i].get() );
#else
      free( s_loop_data->array_1D_Complex[i] );
#endif
   }

   //
   //  De-allocate 2D 7 X loop length Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_2D_7xN_Real_arrays; ++i) {
#if defined(USE_PTR_CLASS)
      free( s_loop_data->array_2D_7xN_Real[i][0].get() );
#else
      free( s_loop_data->array_2D_7xN_Real[i][0] );
#endif
      delete [] s_loop_data->array_2D_7xN_Real[i]; 
   }

   //
   //  De-allocate 2D 64 X 64 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_2D_64x64_Real_arrays; ++i) {
#if defined(USE_PTR_CLASS)
      free( s_loop_data->array_2D_64x64_Real[i][0].get() );
#else
      free( s_loop_data->array_2D_64x64_Real[i][0] );
#endif
      delete [] s_loop_data->array_2D_64x64_Real[i]; 
   }

   //
   //  De-allocate and initialize 3D 2 X loop length X 4 Real arrays.
   //
   for (unsigned i = 0; i < s_loop_data->s_num_3D_2xNx4_Real_arrays; ++i) {
#if defined(USE_PTR_CLASS)
      free( s_loop_data->array_3D_2xNx4_Real[i][0][0].get() );
#else
      free( s_loop_data->array_3D_2xNx4_Real[i][0][0] );
#endif
      for (Indx_type k = 0; k < 2; ++k) {
         delete [] s_loop_data->array_3D_2xNx4_Real[i][k]; 
      }
      delete [] s_loop_data->array_3D_2xNx4_Real[i]; 
   }

   delete s_loop_data;
   s_loop_data = 0;
}


//
// Prototypes for file scope routines used in to manage loop data and checksums
//

namespace {


//
// Routines to allocate and initialize individual arrays consistently for 
// checking results.
//
Real_ptr allocAndInitData(LoopData::RealArray& ra, Indx_type len)
{
   Real_ptr data = 0; 
   posix_memalign( (void **)&data, LCALS_DATA_ALIGN, len*sizeof(Real_type) );
   ra.data = data;
   ra.len = len; 

   initData(ra);

   return data;
}

Indx_type* allocAndInitData(LoopData::IndxArray& ia, Indx_type len)
{
   Indx_type* data = 0;
   posix_memalign( (void **)&data, LCALS_DATA_ALIGN, len*sizeof(Indx_type) );
   ia.data = data;
   ia.len = len;

   initData(ia);

   return data;
}

Complex_ptr allocAndInitData(LoopData::ComplexArray& ca, Indx_type len)
{
   Complex_ptr data = new Complex_type[len];
   ca.data = data; 
   ca.len = len;

   initData(ca);

   return data;
}

void initData(LoopData::RealArray& ra)
{
   int id = ra.id;
   Real_type factor = ( id % 2 ? 0.1 : 0.2 );
   Real_ptr data = ra.data;
   Indx_type len = ra.len;
   for (Indx_type j = 0; j < len; ++j) {
      data[j] = factor*(j + 1.1)/(j + 1.12345);
   }
}

void initData(LoopData::IndxArray& ia)
{
   int id = ia.id;
   Indx_type* data = ia.data;
   Indx_type len = ia.len;
   for (Indx_type j = 0; j < len; ++j) {
      data[j] = 0;
   }
}

void initData(LoopData::ComplexArray& ca)
{
   int id = ca.id;
   Complex_type factor = ( id % 2 ? Complex_type(0.1,0.2) : 
                                    Complex_type(0.2,0.3) );
   Complex_ptr data = ca.data;
   Indx_type len = ca.len;
   for (Indx_type j = 0; j < len; ++j) {
      data[j] = factor*(j + 1.1)/(j + 1.12345);
   }
}


//
// Routines to initialize loop check sum.
//
void initChksum(LoopStat& stat, LoopLength ilength)
{
   stat.loop_chksum[ilength] = 0.0;
}

//
// Routines to update loop check sum.
//
void updateChksum(LoopStat& stat, LoopLength ilength,
                  const LoopData::RealArray& ra,
                  Real_type scale_factor)
{
   Real_ptr data = ra.data;
   Indx_type len = ra.len;
   long double tchk = stat.loop_chksum[ilength];
   for (Indx_type j = 0; j < len; ++j) {
      tchk += (j+1)*data[j]*scale_factor;
   }
   stat.loop_chksum[ilength] = tchk;
}

void updateChksum(LoopStat& stat, LoopLength ilength,
                  Real_type val)
{
   stat.loop_chksum[ilength] += val;
}

void updateChksum(LoopStat& stat, LoopLength ilength,
                  const LoopData::ComplexArray& ca,
                  Real_type scale_factor)
{
   Complex_ptr data = ca.data;
   Indx_type len = ca.len;
   long double tchk = stat.loop_chksum[ilength];
   for (Indx_type j = 0; j < len; ++j) {
      tchk += (j+1)*(real(data[j])+imag(data[j]))*scale_factor;
   }
   stat.loop_chksum[ilength] = tchk;
}

}  // closing brace for unnamed namespace



//
// Recursively construct directories for given path name.
//
bool recursiveMkdir(const std::string& path)
{
   bool retval = true;

   mode_t mode = (S_IRUSR | S_IWUSR | S_IXUSR);
   const char separator = '/';

   int length = static_cast<int>(path.length());
   char* path_buf = new char[length + 1];
   sprintf(path_buf, "%s", path.c_str());
   struct stat status;
   int pos = length - 1;

   /* find part of path that has not yet been created */
   while ((stat(path_buf, &status) != 0) && (pos >= 0)) {

      /* slide backwards in string until next slash found */
      bool slash_found = false;
      while ((!slash_found) && (pos >= 0)) {
         if (path_buf[pos] == separator) {
            slash_found = true;
            if (pos >= 0) path_buf[pos] = '\0';
         } else pos--;
      }
   }

   /*
    * if there is a part of the path that already exists make sure
    * it is really a directory
    */
   if (pos >= 0) {
      if (!S_ISDIR(status.st_mode)) {
         std::cout << "Cannot create directories in path = " << path
                   << "\n    because some intermediate item in path exists and"
                   << "is NOT a directory" << std::endl;
         retval = false;
      }
   }

   /*
    * make all directories that do not already exist
    *
    * if (pos < 0), then there is no part of the path that
    * already exists.  Need to make the first part of the
    * path before sliding along path_buf.
    */
   if ( retval && pos < 0) {
      if (mkdir(path_buf, mode) != 0) {
         std::cout << "   Cannot create directory  = "
                   << path_buf << std::endl;
         retval = false;
      }
      pos = 0;
   }

   if ( retval ) {

      /* make remaining directories */
      do {

         /* slide forward in string until next '\0' found */
            bool null_found = false;
         while ((!null_found) && (pos < length)) {
            if (path_buf[pos] == '\0') {
               null_found = true;
            path_buf[pos] = separator;
            }
            pos++;
         }

         /* make directory if not at end of path */
         if (pos < length) {
            if (mkdir(path_buf, mode) != 0) {
              std::cout << "   Cannot create directory  = "
                        << path_buf << std::endl;
              retval = false;
            }
         }
      } while (pos < length && retval);

   }

   delete[] path_buf;

   return retval;
}
