//
// See README-LCALS_license.txt for access and distribution restrictions
//

//
// Header file with enums, macros, routines and structures used to
// compile and run loops in LCALS suite and to generate execution
// statistics.
//

#ifndef LCALSSuite_HXX
#define LCALSSuite_HXX

#include "LCALSParams.hxx" 
#include "LCALSStats.hxx"

#include <vector>
#include <string>

//
// Enumeration defining unique id for each loop in suite
//
// Note: to keep output understandable, keep this consistent with
//       routine defineLoopSuiteRunInfo().
//
enum LoopID {

   // Keep this one first and don't comment out (!!)
   // This insures loop ids start at zero so all array indexing
   // or data structures is correct.  Also, this loop is not
   // executed the same way the others are. 
   REF_LOOP = 0,

   //
   // Loop Subset A: Loops extracted from LLNL app codes.
   // They are implemented in runA<variant>Loops.cxx files.
   //
   PRESSURE_CALC,
   ENERGY_CALC,
   VOL3D_CALC,
   DEL_DOT_VEC_2D,
   COUPLE,
   FIR,

   //
   // Loop Subset B: "Basic" Loops.
   // They are implemented in runB<variant>Loops.cxx files.
   //
   INIT3,
   MULADDSUB,
   IF_QUAD,
   TRAP_INT,

   //
   // Loop Subset C: Loops from older Livermore Loops in "C" suite.
   // They are implemented in runC<variant>Loops.cxx files.
   //
   HYDRO_1D,
   ICCG,
   INNER_PROD,
   BAND_LIN_EQ,
   TRIDIAG_ELIM,
   EOS,
   ADI,
   INT_PREDICT,
   DIFF_PREDICT,
   FIRST_SUM,
   FIRST_DIFF,
   PIC_2D,
   PIC_1D,
   HYDRO_2D,
   GEN_LIN_RECUR,
   DISC_ORD,
   MAT_X_MAT,
   PLANCKIAN,
   IMP_HYDRO_2D,
   FIND_FIRST_MIN,

   NUM_LOOPS // Keep this one last and NEVER comment out (!!)

};


//
// Enumeration defining possible loop lengths to run.
//
enum LoopLength {

   LONG = 0,
   MEDIUM,
   SHORT,

   NUM_LENGTHS // Keep this one last (!!)

};


//
// The following macro constants are used to turn on/off compilation of
// individual loops in suite.  Names are consistent with LoopID enum above.
//

//#define DO_OMP_ONLY
#undef DO_OMP_ONLY

#if defined (DO_OMP_ONLY)

#define COMPILE_PRESSURE_CALC
#define COMPILE_ENERGY_CALC
#define COMPILE_VOL3D_CALC
#define COMPILE_DEL_DOT_VEC_2D
#define COMPILE_COUPLE
#define COMPILE_FIR
#define COMPILE_IF_QUAD
#define COMPILE_TRAP_INT
#define COMPILE_PIC_2D

#else
//
// Loop Subset A: Loops extracted from LLNL app codes.
// They are implemented in runA<variant>Loops.cxx files.
//
#define COMPILE_PRESSURE_CALC
#define COMPILE_ENERGY_CALC
#define COMPILE_VOL3D_CALC
#define COMPILE_DEL_DOT_VEC_2D
#define COMPILE_COUPLE
#define COMPILE_FIR

//
// Loop Subset B: "Basic" Loops.
// They are implemented in runB<variant>Loops.cxx files.
//
#define COMPILE_INIT3
#define COMPILE_MULADDSUB
#define COMPILE_IF_QUAD
#define COMPILE_TRAP_INT

//
// Loop Subset C: Loops from older Livermore Loops in "C" suite.
// They are implemented in runLCK<variant>Loops.cxx files.
//
#define COMPILE_HYDRO_1D
#define COMPILE_ICCG
#define COMPILE_INNER_PROD
#define COMPILE_BAND_LIN_EQ
#define COMPILE_TRIDIAG_ELIM
#define COMPILE_EOS
#define COMPILE_ADI
#define COMPILE_INT_PREDICT
#define COMPILE_DIFF_PREDICT
#define COMPILE_FIRST_SUM
#define COMPILE_FIRST_DIFF
#define COMPILE_PIC_2D
#define COMPILE_PIC_1D
#define COMPILE_HYDRO_2D
#define COMPILE_GEN_LIN_RECUR
#define COMPILE_DISC_ORD
#define COMPILE_MAT_X_MAT
#define COMPILE_PLANCKIAN
#define COMPILE_IMP_HYDRO_2D
#define COMPILE_FIND_FIRST_MIN
#endif



//////////////////////////////////////////////////////////////////
//
//  Structure holding double arrays and scalars used in loops.
//
//  Note: These are initialized in allocateLoopData().
//
///////////////////////////////////////////////////////////////////

struct LoopData
{

   //
   // Structures to hold data for easy reinitialization
   // (useful for verifying result checksums, etc.)
   //  
   struct RealArray 
   {
      int id;
      Real_ptr data; 
      Indx_type len;
   };

   struct IndxArray
   {
      int id;
      Indx_type* data;
      Indx_type  len;
   };

   struct ComplexArray
   {
      int id;
      Complex_ptr data;
      Indx_type  len;
   };


   Indx_type max_loop_length;

   //
   // Static values indicating number of data arrays 
   // of various forms used in loop suite.
   //
   // NOTE: These number may need to change to accomodate new loops. 
   //       Also, other arrays may need to be added. 
   //
   static const unsigned s_num_1D_Real_arrays = 16;
   static const unsigned s_num_1D_Nx4_Real_arrays = 2;
   static const unsigned s_num_1D_Indx_arrays = 5;
   static const unsigned s_num_1D_Complex_arrays = 5;

   static const unsigned s_num_2D_Nx25_Real_arrays = 4;
   static const unsigned s_num_2D_7xN_Real_arrays = 11;
   static const unsigned s_num_2D_64x64_Real_arrays = 1;

   static const unsigned s_num_3D_2xNx4_Real_arrays = 3;

   static const unsigned s_num_Real_scalars   = 10;

   //
   // NOTE: To see how the following data structures are related,
   //       please see the routine allocateLoopData() in the 
   //       file LCALSSuite.cxx.
   //
   //       The reason that we hold on to the same data in two 
   //       different ways is two-fold:
   //          1) The first set of arrays below makes it easy to 
   //             access pointers to data based on what is used in 
   //             each loop kernel; e.g., arrays of variaous dimensions.
   //          2) The second set of arrays makes it easy to process
   //             arrays for (re)initialization and checksum 
   //             computation to verify results; e.g., we simply
   //             iterate through 1-dim arrays without having to
   //             know their lengths, if they are really being used 
   //             as 2- or 3-dimensional arrays, for example. 
   //

   //
   // Data arrays and scalars used in loop execution.
   //
   Real_ptr  array_1D_Real[s_num_1D_Real_arrays];
   Real_ptr  array_1D_Nx4_Real[s_num_1D_Nx4_Real_arrays];
   Indx_type* array_1D_Indx[s_num_1D_Indx_arrays]; 
   Complex_ptr array_1D_Complex[s_num_1D_Complex_arrays]; 

   Real_ptr* array_2D_Nx25_Real[s_num_2D_Nx25_Real_arrays];
   Real_ptr* array_2D_7xN_Real[s_num_2D_7xN_Real_arrays];
   Real_ptr* array_2D_64x64_Real[s_num_2D_64x64_Real_arrays];

   Real_ptr** array_3D_2xNx4_Real[s_num_3D_2xNx4_Real_arrays];

   Real_type scalar_Real[s_num_Real_scalars];

   //
   // Arrays of structs holding data arrays used for data initialization
   // and checksum verification.
   //
   RealArray RealArray_1D[s_num_1D_Real_arrays];
   RealArray RealArray_1D_Nx4[s_num_1D_Nx4_Real_arrays];
   IndxArray IndxArray_1D[s_num_1D_Indx_arrays];
   ComplexArray ComplexArray_1D[s_num_1D_Complex_arrays];

   RealArray RealArray_2D_Nx25[s_num_2D_Nx25_Real_arrays];
   RealArray RealArray_2D_7xN[s_num_2D_7xN_Real_arrays];
   RealArray RealArray_2D_64x64[s_num_2D_64x64_Real_arrays];

   RealArray RealArray_3D_2xNx4[s_num_3D_2xNx4_Real_arrays];

   RealArray RealArray_scalars;

};

//
// Routine to access data structure that holds data needed to execute loops.
// 
LoopData& getLoopData();



//////////////////////////////////////////////////////////////////
//
//  Routines to define how loop suite will be run and
//  to set up data for loop suite.
//
//////////////////////////////////////////////////////////////////

//
// Routines to define specific details about how to run loop suite.
//
// Note:  Individual loop lengths and sampling parameters 
//        are defined in this routine.
//
void defineLoopSuiteRunInfo(bool run_loop[],
                            std::vector< std::string >& run_loop_variants,
                            double sample_frac);


//
// Routines to allocate and initialize arrays (and scalars) for 
// loops in suite and to free those arrays when done.
//
void allocateLoopData();
void freeLoopData();


//
// Routines to initialize and finalize loop data, statistics, timers, etc.  
//
// Each of these routines must be called before and after the execution
// of each loop.
//
void loopInit(unsigned iloop, LoopStat& stat);
//
void loopFinalize(unsigned iloop, LoopStat& stat, LoopLength ilength);


//
// Routines to run reference loops for figure of merit (FOM) calculations.
//
void defineReferenceLoopRunInfo();
void computeReferenceLoopTimes();


//
// Routines to run loop variants for suite.
//
// loop_stats is vector of LoopStat objects corresponding to loop variant.
// run_loop boolean array indicates which loops in suite to execute 
// ilength indicates which loop length to run (see LoopLength enum).
//
void runARawLoops( std::vector<LoopStat>& loop_stats,
                   bool run_loop[],
                   LoopLength ilength );
void runBRawLoops( std::vector<LoopStat>& loop_stats,
                   bool run_loop[],
                   LoopLength ilength );
void runCRawLoops( std::vector<LoopStat>& loop_stats,
                   bool run_loop[],
                   LoopLength ilength );

void runARawFuncLoops( std::vector<LoopStat>& loop_stats,
                       bool run_loop[],
                       LoopLength ilength );
void runBRawFuncLoops( std::vector<LoopStat>& loop_stats,
                       bool run_loop[],
                       LoopLength ilength );
void runCRawFuncLoops( std::vector<LoopStat>& loop_stats,
                       bool run_loop[],
                       LoopLength ilength );

void runOMPRawLoops( std::vector<LoopStat>& loop_stats,
                     bool run_loop[],
                     LoopLength ilength );


void runAForallLambdaLoops( std::vector<LoopStat>& loop_stats,
                            bool run_loop[],
                            LoopLength ilength );
void runBForallLambdaLoops( std::vector<LoopStat>& loop_stats,
                            bool run_loop[],
                            LoopLength ilength );
void runCForallLambdaLoops( std::vector<LoopStat>& loop_stats,
                            bool run_loop[],
                            LoopLength ilength );
void runOMPForallLambdaLoops( std::vector<LoopStat>& loop_stats,
                              bool run_loop[],
                              LoopLength ilength );

void runAForallLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                    bool run_loop[],
                                    LoopLength ilength );
void runBForallLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                    bool run_loop[],
                                    LoopLength ilength );
void runCForallLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                    bool run_loop[],
                                    LoopLength ilength );
void runOMPForallLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                      bool run_loop[],
                                      LoopLength ilength );


void runAForallFunctorLoops( std::vector<LoopStat>& loop_stats,
                             bool run_loop[],
                             LoopLength ilength );
void runBForallFunctorLoops( std::vector<LoopStat>& loop_stats,
                             bool run_loop[],
                             LoopLength ilength );
void runCForallFunctorLoops( std::vector<LoopStat>& loop_stats,
                             bool run_loop[],
                             LoopLength ilength );
void runOMPForallFunctorLoops( std::vector<LoopStat>& loop_stats,
                               bool run_loop[],
                               LoopLength ilength );

void runAForallHybridLambdaLoops( std::vector<LoopStat>& loop_stats,
                                  bool run_loop[],
                                  LoopLength ilength );
void runBForallHybridLambdaLoops( std::vector<LoopStat>& loop_stats,
                                  bool run_loop[],
                                  LoopLength ilength );
void runCForallHybridLambdaLoops( std::vector<LoopStat>& loop_stats,
                                  bool run_loop[],
                                  LoopLength ilength );

void runAForallHybridLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                          bool run_loop[],
                                          LoopLength ilength );
void runBForallHybridLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                          bool run_loop[],
                                          LoopLength ilength );
void runCForallHybridLambdaLoops_TYPEFIX( std::vector<LoopStat>& loop_stats,
                                          bool run_loop[],
                                          LoopLength ilength );

//
// Recursively construct directories based on a relative or
// absolute path name.  Return true if directory created
// successfully, else false.
// 
bool recursiveMkdir(const std::string& path);




#endif  // closing endif for header file include guard
