//
// See README-LCALS_license.txt for access and distribution restrictions
//

//
// Header file containing LCALS traversal method templates used with
// "forall" loop variants.
//
// Tag structs for traversal types are located in LCALSParams.hxx. 
//

#ifndef LCALSTraversalMethods_HXX
#define LCALSTraversalMethods_HXX

#include "LCALSParams.hxx"

#include <vector>


/*!
 ******************************************************************************
 *
 * \brief  Traverse contiguous range of indices using sequential execution.
 *
 ******************************************************************************
 */
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(seq_traversal,
            Indx_type begin, Indx_type end, LOOP_BODY loop_body)
{
#pragma novector
   for ( Indx_type ii = begin ; ii < end ; ++ii ) {
      loop_body( ii );
   }
}

/// with stride
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(seq_traversal,
            Indx_type begin, Indx_type end, Indx_type stride,
            LOOP_BODY loop_body)
{
#pragma novector
   for ( Indx_type ii = begin ; ii < end ; ii += stride ) {
      loop_body( ii );
   }
}


/*!
 ******************************************************************************
 *
 * \brief  Traverse contiguous range of indices using SIMD vectorization.
 *         No assumption made on data alignment.
 *
 ******************************************************************************
 */
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(simd_traversal,
            Indx_type begin, Indx_type end, LOOP_BODY loop_body)
{
   for ( Indx_type ii = begin ; ii < end ; ++ii ) {
      loop_body( ii );
   }
}

/// with stride
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(simd_traversal,
            Indx_type begin, Indx_type end, Indx_type stride,
            LOOP_BODY loop_body)
{
   for ( Indx_type ii = begin ; ii < end ; ii += stride ) {
      loop_body( ii );
   }
}


/*!
 ******************************************************************************
 *
 * \brief  Traverse contiguous range of indices using OpenMP parallel for.
 *
 ******************************************************************************
 */
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(omp_parallel_for_traversal,
            Indx_type begin, Indx_type end, LOOP_BODY loop_body)
{
#pragma omp parallel for schedule(static)
   for ( Indx_type ii = begin ; ii < end ; ++ii ) {
      loop_body( ii );
   }
}

/// with stride
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(omp_parallel_for_traversal,
            Indx_type begin, Indx_type end, Indx_type stride,
            LOOP_BODY loop_body)
{
#pragma omp parallel for schedule(static)
   for ( Indx_type ii = begin ; ii < end ; ii += stride ) {
      loop_body( ii );
   }
}


/*!
 ******************************************************************************
 *
 * \brief  Traverse contiguous range of indices using OpenMP for with
 *         nowait clause.
 *
 ******************************************************************************
 */
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(omp_for_nowait_traversal,
            Indx_type begin, Indx_type end, LOOP_BODY loop_body)
{
#pragma omp for nowait schedule(static)
   for ( Indx_type ii = begin ; ii < end ; ++ii ) {
      loop_body( ii );
   }
}

/// with stride
template <typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(omp_for_nowait_traversal,
            Indx_type begin, Indx_type end, Indx_type stride,
            LOOP_BODY loop_body)
{
#pragma omp for nowait schedule(static)
   for ( Indx_type ii = begin ; ii < end ; ii += stride ) {
      loop_body( ii );
   }
}


/*!
 ******************************************************************************
 *
 * \brief Generic methods with traversal policy specified by template 
 *        parameter.
 *
 ******************************************************************************
 */
template <typename TRAVERSE_T, typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(Indx_type begin, Indx_type end, LOOP_BODY loop_body)
{
   forall( TRAVERSE_T(), begin, end, loop_body );
}

/// with stride
template <typename TRAVERSE_T, typename LOOP_BODY>
INLINE_DIRECTIVE
void forall(Indx_type begin, Indx_type end, Indx_type stride,
            LOOP_BODY loop_body)
{
   forall( TRAVERSE_T(), begin, end, stride, loop_body );
}


#endif  // closing endif for header file include guard
