//
// See README-LCALS_license.txt for access and distribution restrictions
//
// This code is under continuing development.  Go to http://codesign.llnl.gov
// to acquire the latest released version.
//

//
// Main program for LCALS suite.
//

#include <cstdlib>

#include<string>
#include<iostream>
#include<algorithm>

#include <unistd.h>

#include "LCALSSuite.hxx"


int main(int argc, char *argv[])
{

   //
   // Process command line args
   //
   bool do_fom = false;
   bool input_error = false;
   std::string output_dirname;

   // if (argc == 1) no args to check...

   if (argc > 1) {
      // check whether FOM option is given
      std::string arg = argv[1];
      if ( arg == "-fom" ) {
         do_fom = true; 
      } else {
         output_dirname = argv[1];
      }
   }

   if (argc > 2) {
      if ( !do_fom ) { 
         input_error = true; 
      } else {
         output_dirname = argv[2];
      }
   }

   if ( argc > 3) {
      input_error = true; 
   } 

   if ( !input_error ) { 

      if ( !output_dirname.empty() && !recursiveMkdir(output_dirname) ) {
        std::cout << "Problem with given output directory name." << std::endl;
        std::cout << "No file output will be generated." << std::endl;
      } 

   } else {
      std::cout << "USAGE:  " << argv[0] 
                << "-fom <output directory name>, both optional" << std::endl;
      std::cout << "\tIf '-fom' option is given, " 
                << "FOM will be computed and reported." << std::endl;
      std::cout << "Run summary will be printed to standard output, " 
                << "if no ouytput directory name is given." << std::endl;
      std::cout << "\tIf directory name is provided, execution summary and "
                << "text files suitable for importing into MS Excel will be written there." << std::endl;
      exit(-1);
      return -1;
   }


   //
   // Define some parameters specifying how suite of loops will execute.
   //
   // See README-LCALS_instructions.txt file for additional description of how 
   // to control compilation and execution of loop suite.
   //
   unsigned num_suite_passes = 1;
#if defined(VERIFY_CHECKSUM_ABBREVIATED)
   //
   // When verifying checksums, we only take one pass through the suite of loops
   // as this is sufficient.
   // 
   num_suite_passes = num_checksum_suite_passes;
#endif

   //
   // Specify fraction of pre-defined loop sample counts to use.
   // Smaller value reduces totoal run time.  However, a value too
   // small will result in inaccurate timings.
   //
   double sample_frac = 1.0;

   //
   // Specify which loop variants are executed by adding appropriate
   // string name to 'run_loop_variants' array.  
   //
   // NOTE: First one pushed onto array will be reference variant
   //       to which others are compared in run time statistics.
   //
   std::vector< std::string > run_loop_variants;
#if defined(DO_RAW)
   run_loop_variants.push_back("Raw"); 
#endif
#if defined(DO_RAW_OMP)
   run_loop_variants.push_back("Raw_OMP"); 
#endif

#if defined(DO_LAMBDA)
   run_loop_variants.push_back("Forall_Lambda"); 
#endif
#if defined(DO_LAMBDA_OMP)
   run_loop_variants.push_back("Forall_Lambda_OMP"); 
#endif

   //
   //  Specify which loops lengths to run by true/false
   //  value in 'run_loop_length' array.
   //
   bool run_loop_length[NUM_LENGTHS];
   run_loop_length[LONG] = true;
   run_loop_length[MEDIUM] = true;
   run_loop_length[SHORT] = true;


   //
   //  Specify which loops to run by true/false value in 'run_loop' array.
   //
   bool run_loop[NUM_LOOPS];
   for (unsigned iloop = 0; iloop < NUM_LOOPS; ++iloop) run_loop[iloop] = false;

   // Loop Subset A: Loops extracted from LLNL app codes.
   run_loop[PRESSURE_CALC ] = true;
   run_loop[ENERGY_CALC   ] = true;
   run_loop[VOL3D_CALC    ] = true;
   run_loop[DEL_DOT_VEC_2D] = true;
   run_loop[COUPLE        ] = true;
   run_loop[FIR           ] = true;

   // Loop Subset B: "Basic" Loops.
   run_loop[INIT3         ] = true;
   run_loop[MULADDSUB     ] = true;
   run_loop[IF_QUAD       ] = true;
   run_loop[TRAP_INT      ] = true;

   // Loop Subset C: Loops from older Livermore Loops in "C" suite.
   run_loop[HYDRO_1D      ] = true;
   run_loop[ICCG          ] = true;
   run_loop[INNER_PROD    ] = true;
   run_loop[BAND_LIN_EQ   ] = true;
   run_loop[TRIDIAG_ELIM  ] = true;
   run_loop[EOS           ] = true;
   run_loop[ADI           ] = true;
   run_loop[INT_PREDICT   ] = true; 
   run_loop[DIFF_PREDICT  ] = true; 
   run_loop[FIRST_SUM     ] = true;
   run_loop[FIRST_DIFF    ] = true;
   run_loop[PIC_2D        ] = true; 
   run_loop[PIC_1D        ] = true;
   run_loop[HYDRO_2D      ] = true;
   run_loop[GEN_LIN_RECUR ] = true;
   run_loop[DISC_ORD      ] = true;
   run_loop[MAT_X_MAT     ] = true;
   run_loop[PLANCKIAN     ] = true;
   run_loop[IMP_HYDRO_2D  ] = true;
   run_loop[FIND_FIRST_MIN] = true;


   //
   // Obtain and report hostname.
   // 
   const int host_namelen = 64;
   char host[host_namelen];
   gethostname( host, host_namelen );
   std::string host_name(host);

   std::cout << "\n Running loop suite on " << host_name << std::endl;
   
   //
   // Specify size in bytes of largest data cache level on machine so that 
   // caches can be properly flushed between execution of different loops.
   // 
   CacheIndx_type cache_size = 0;
   if ( host_name.find("rzalastor") != std::string::npos ) {
      cache_size = 12000000;  // 12MB on rzalastor 
   } else if ( host_name.find("rzmerl") != std::string::npos ) {
      cache_size = 20000000;  // 20MB on rzmerl  
   } else if ( host_name.find("dawn") != std::string::npos ) {
      cache_size = 8000000;   // 8MB on dawn/rzdawndev
   } else if ( host_name.find("rzuseq") != std::string::npos ||
               host_name.find("vulcan") != std::string::npos ||
               host_name.find("sequoia") != std::string::npos ) {
      cache_size = 32000000;  // 32MB on BG/Q
   } else {
      std::cout << "\n WARNING: unknown system cache size. " 
                << "Timing results may be suspect!!" << std::endl;
   }

 
   //
   // Allocate data for running loops and generating execution timings.
   // Also, set structures that define how loops will be run.
   //
   allocateLoopSuiteRunInfo(host_name,
                            NUM_LOOPS,
                            NUM_LENGTHS,
                            num_suite_passes, 
                            run_loop_length,
                            cache_size);

   defineLoopSuiteRunInfo(run_loop, run_loop_variants, sample_frac);

   allocateLoopData();


   if (do_fom) {
      //
      // Compute reference times for figure of merit (FOM) calculation.
      //
      computeReferenceLoopTimes();
   }


   //
   // Run loops.
   //
   LoopSuiteRunInfo& loop_suite_run_info = getLoopSuiteRunInfo();

   for (unsigned ipass = 0; ipass < num_suite_passes; ++ipass) {
      std::cout << "\n run suite: pass = " << ipass << std::endl;

      for (unsigned ivariant = 0; 
           ivariant < run_loop_variants.size(); ++ivariant) {

         std::string& loop_variant = run_loop_variants[ivariant];
         std::vector<LoopStat>& loop_stats =
              loop_suite_run_info.getLoopStats(loop_variant);

         std::cout << "\t run loop variant ---> " << loop_variant << std::endl;

         for (unsigned ilen = 0; ilen < NUM_LENGTHS; ++ilen) {

            if (run_loop_length[ilen]) {

               LoopLength rilen = static_cast<LoopLength>(ilen);

               if ( loop_variant == "Raw" ) {

#if defined(DO_RAW)
                  runARawLoops(loop_stats, run_loop, rilen);
                  runBRawLoops(loop_stats, run_loop, rilen);
                  runCRawLoops(loop_stats, run_loop, rilen);
#endif

               } else if ( loop_variant == "Raw_OMP" ) {

#if defined(DO_RAW_OMP)
                  runOMPRawLoops(loop_stats, run_loop, rilen);
#endif

               } else if ( loop_variant == "Forall_Lambda" ) {

#if defined(DO_LAMBDA)
                  runAForallLambdaLoops(loop_stats, run_loop, rilen);
                  runBForallLambdaLoops(loop_stats, run_loop, rilen);
                  runCForallLambdaLoops(loop_stats, run_loop, rilen);
#endif

               } else if ( loop_variant == "Forall_Lambda_OMP" ) {

#if defined(DO_LAMBDA_OMP)
                  runOMPForallLambdaLoops(loop_stats, run_loop, rilen);
#endif

               } else {
                  std::cout << "\n Unknown loop variant = " 
                            << loop_variant << std::endl;
               }
 
            }  // if loop length is run

         } // iterate over loop lengths

      }  // iterate over loop variants

   }  // iterate over loop suite passes


   //
   //  Generate report(s).
   //
   std::cout << "\n generate reports...." << std::endl;

   generateTimingReport(run_loop_variants, output_dirname, do_fom);
   generateChecksumReport(run_loop_variants, output_dirname); 
   if (do_fom) {
      generateFOMReport(run_loop_variants, output_dirname);
   }


   //
   //  Clean up.
   //
   freeLoopData();

   std::cout << "\n freeLoopSuiteRunInfo..." << std::endl;
   freeLoopSuiteRunInfo();

   std::cout << "\n DONE!!! " << std::endl;

   return 0 ;
}

